@@ -70,10 +70,12 @@ struct wl_proxy {
 	wl_dispatcher_func_t dispatcher;
 	uint32_t version;
 	const char * const *tag;
+	struct wl_list queue_link; /**< in struct wl_event_queue::proxy_list */
 };
 
 struct wl_event_queue {
 	struct wl_list event_list;
+	struct wl_list proxy_list; /**< struct wl_proxy::queue_link */
 	struct wl_display *display;
 };
 
@@ -221,6 +223,7 @@ static void
 wl_event_queue_init(struct wl_event_queue *queue, struct wl_display *display)
 {
 	wl_list_init(&queue->event_list);
+	wl_list_init(&queue->proxy_list);
 	queue->display = display;
 }
 
@@ -298,6 +301,27 @@ wl_event_queue_release(struct wl_event_queue *queue)
 {
 	struct wl_closure *closure;
 
+	if (!wl_list_empty(&queue->proxy_list)) {
+		struct wl_proxy *proxy, *tmp;
+
+		if (queue != &queue->display->default_queue) {
+			wl_log("warning: queue %p destroyed while proxies "
+			       "still attached:\n", queue);
+		}
+
+		wl_list_for_each_safe(proxy, tmp, &queue->proxy_list,
+				      queue_link) {
+			if (queue != &queue->display->default_queue) {
+				wl_log("  %s@%u still attached\n",
+				       proxy->object.interface->name,
+				       proxy->object.id);
+			}
+			proxy->queue = NULL;
+			wl_list_remove(&proxy->queue_link);
+			wl_list_init(&proxy->queue_link);
+		}
+	}
+
 	while (!wl_list_empty(&queue->event_list)) {
 		closure = wl_container_of(queue->event_list.next,
 					  closure, link);
@@ -435,6 +459,8 @@ proxy_create(struct wl_proxy *factory, const struct wl_interface *interface,
 		return NULL;
 	}
 
+	wl_list_insert(&proxy->queue->proxy_list, &proxy->queue_link);
+
 	return proxy;
 }
 
@@ -494,6 +520,8 @@ wl_proxy_create_for_id(struct wl_proxy *factory,
 		return NULL;
 	}
 
+	wl_list_insert(&proxy->queue->proxy_list, &proxy->queue_link);
+
 	return proxy;
 }
 
@@ -518,6 +546,10 @@ proxy_destroy(struct wl_proxy *proxy)
 
 	proxy->flags |= WL_PROXY_FLAG_DESTROYED;
 
+	proxy->queue = NULL;
+	wl_list_remove(&proxy->queue_link);
+	wl_list_init(&proxy->queue_link);
+
 	wl_proxy_unref(proxy);
 }
 
@@ -1276,8 +1308,9 @@ wl_display_connect(const char *name)
  *
  * \param display The display context object
  *
- * Close the connection to \c display and free all resources associated
- * with it.
+ * Close the connection to \c display. The \ref wl_proxy and
+ * \ref wl_event_queue objects need to be manually destroyed by the caller
+ * before disconnecting.
  *
  * \memberof wl_display
  */
@@ -1538,6 +1571,9 @@ queue_event(struct wl_display *display, int len)
 	else
 		queue = proxy->queue;
 
+	if (!queue)
+		wl_abort("Tried to add event to destroyed queue\n");
+
 	wl_list_insert(queue->event_list.prev, &closure->link);
 
 	return size;
@@ -2335,6 +2371,8 @@ wl_proxy_set_queue(struct wl_proxy *proxy, struct wl_event_queue *queue)
 {
 	pthread_mutex_lock(&proxy->display->mutex);
 
+	wl_list_remove(&proxy->queue_link);
+
 	if (queue) {
 		assert(proxy->display == queue->display);
 		proxy->queue = queue;
@@ -2342,6 +2380,8 @@ wl_proxy_set_queue(struct wl_proxy *proxy, struct wl_event_queue *queue)
 		proxy->queue = &proxy->display->default_queue;
 	}
 
+	wl_list_insert(&proxy->queue->proxy_list, &proxy->queue_link);
+
 	pthread_mutex_unlock(&proxy->display->mutex);
 }
 
@@ -2413,6 +2453,8 @@ wl_proxy_create_wrapper(void *proxy)
 	wrapper->flags = WL_PROXY_FLAG_WRAPPER;
 	wrapper->refcount = 1;
 
+	wl_list_insert(&wrapper->queue->proxy_list, &wrapper->queue_link);
+
 	pthread_mutex_unlock(&wrapped_proxy->display->mutex);
 
 	return wrapper;
@@ -2434,6 +2476,12 @@ wl_proxy_wrapper_destroy(void *proxy_wrapper)
 
 	assert(wrapper->refcount == 1);
 
+	pthread_mutex_lock(&wrapper->display->mutex);
+
+	wl_list_remove(&wrapper->queue_link);
+
+	pthread_mutex_unlock(&wrapper->display->mutex);
+
 	free(wrapper);
 }
 
